import torch
from typing import Callable, Iterable, overload, TypeVar, Union

from .api.types import Minibatches, MinibatchTransform

def stop_grad(fn: Callable, *args, **kwargs):
    with torch.no_grad():
        return fn(*args, **kwargs)

def no_grad(x: Iterable):
    with torch.no_grad():
        yield from x

T = TypeVar("T")

@overload
def pipe(
    tensor_or_minibatches: "torch.Tensor",
    immediate_call: Callable[["torch.Tensor"], T],
    minibatch_call: MinibatchTransform
) -> T:
    ...
@overload
def pipe(
    tensor_or_minibatches: Minibatches,
    immediate_call: Callable[["torch.Tensor"], T],
    minibatch_call: MinibatchTransform
) -> Minibatches:
    ...
@overload
def pipe(
    tensor_or_minibatches: Union["torch.Tensor", Minibatches],
    immediate_call: Callable[["torch.Tensor"], T],
    minibatch_call: MinibatchTransform
) -> Union[T, Minibatches]:
    ...
def pipe(
    tensor_or_minibatches: Union["torch.Tensor", Minibatches],
    immediate_call: Callable[["torch.Tensor"], T],
    minibatch_call: MinibatchTransform
) -> Union[T, Minibatches]:
    if isinstance(tensor_or_minibatches, torch.Tensor):
        return immediate_call(tensor_or_minibatches)
    return map(minibatch_call, tensor_or_minibatches)
